"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.NiketaClient = void 0;
const common_1 = require("./utils/common");
const execa_1 = __importDefault(require("execa"));
const fs_1 = require("fs");
const helpers_fn_1 = require("helpers-fn");
const lint_fn_1 = require("lint-fn");
const net_1 = require("net");
const rambdax_1 = require("rambdax");
const lint_only_mode_1 = require("./modules/lint-only-mode");
const create_file_key_1 = require("./utils/create-file-key");
const get_coverage_path_1 = require("./utils/get-coverage-path");
const get_spec_file_1 = require("./utils/get-spec-file");
const get_new_decorations_1 = require("./utils/get-new-decorations");
const get_uncovered_message_1 = require("./utils/get-uncovered-message");
const EXTENDED_LOG = false;
const FUNCTIONS = '☈'; // 🕸
const STALE_SEPARATOR = '☄'; // 🌰
const SEPARATOR = ' '; // 🎑 🔘 🧱
const STATEMENTS = '✍';
const BRANCHES = '🎋';
const LINES = '📜';
function logJest(execResult, enabled) {
    if (!enabled)
        return;
    process.stderr.write('\n🐬\n' + execResult.stderr + '\n🐬\n');
    process.stderr.write('\n🍵\n' + execResult.stdout + '\n🍵\n');
}
function fileInfo(x) {
    const [firstFolder, fileName] = rambdax_1.takeLast(2, x.split('/'));
    return `${firstFolder}/${fileName}`;
}
function onWrongIncomingMessage(message) {
    console.log({
        message: message.toString(),
        type: typeof message,
    });
    return helpers_fn_1.log('Error while parsing messageFromVSCode', 'error');
}
function debugLog(toLog, label = '') {
    if (!EXTENDED_LOG)
        return;
    console.log(label, common_1.SHORT_SEPARATOR);
    console.log(common_1.LONG_SEPARATOR);
    console.log(toLog);
    console.log(common_1.LONG_SEPARATOR);
}
class NiketaClient {
    constructor(input) {
        this.port = input.port;
        this.testing = Boolean(input.testing);
        this.coverageHolder = {};
        this.emit = input.emit === undefined ? common_1.defaultEmit : input.emit;
        this.initialized = false;
    }
    async onJestMessage(message) {
        const { fileName, dir, hasTypescript, requestLintFile } = message;
        if (!common_1.isMessageCorrect(message)) {
            return this.emtpyAnswer(fileName, 'message is incorrect');
        }
        const lintOnly = lint_only_mode_1.isLintOnlyMode(fileName);
        const lintMessage = ` ${fileInfo(fileName)}`;
        /*
          Jest is setup for one of the two so we shouldn't check for both
        */
        const allowedSpecExtension = hasTypescript ? '.ts' : '.js';
        const { hasValidSpec, specFile } = get_spec_file_1.getSpecFile(fileName, allowedSpecExtension);
        debugLog({
            lintOnly,
            hasValidSpec,
        });
        if (requestLintFile) {
            debugLog('requestLintFile');
            return this.handleRequestLint({
                fileName,
                lintOnly,
                lintMessage,
            });
        }
        if (lintOnly) {
            debugLog('lintOnly');
            await lint_only_mode_1.lintOnlyMode(fileName);
            return this.lintAnswer(lintMessage);
        }
        if (!hasValidSpec)
            return this.emtpyAnswer(fileName, 'spec is not valid');
        const [failure, execResult, actualFileName, extension,] = await this.execJest({
            dir,
            fileName: fileName,
            specFileName: specFile,
        });
        if (failure) {
            return this.emtpyAnswer(fileName, 'Jest stopped for known or unknown reasons');
        }
        logJest(execResult, !this.testing);
        return this.onJestSuccess({
            specFile,
            dir,
            execResult,
            actualFileName,
            fileName,
            extension,
            hasTypescript,
        });
    }
    async handleRequestLint(input) {
        const { fileName, lintOnly, lintMessage } = input;
        if (!lintOnly && !common_1.isLintable(fileName)) {
            return this.emtpyAnswer(fileName, '!lintable');
        }
        if (lintOnly) {
            await lint_only_mode_1.lintOnlyMode(fileName);
            return this.lintAnswer(lintMessage);
        }
        try {
            await this.applyLint(fileName);
            return this.lintAnswer(lintMessage);
        }
        catch (_) {
            return this.lintAnswer(`${common_1.ERROR_ICON} ${lintMessage}`);
        }
    }
    emtpyAnswer(fileName, reason) {
        debugLog(reason);
        this.emit({
            firstBarMessage: 'NO ACTION',
            secondBarMessage: undefined,
            thirdBarMessage: `${reason} - ${fileInfo(fileName)}`,
            hasDecorations: false,
        });
    }
    lintAnswer(lintMessage) {
        this.emit({
            firstBarMessage: 'LINT ACTION',
            secondBarMessage: undefined,
            thirdBarMessage: lintMessage,
            hasDecorations: false,
        });
    }
    async applyLint(fileName) {
        if (!fs_1.existsSync(fileName))
            return helpers_fn_1.log(`${fileName} is deleted`, 'error');
        helpers_fn_1.log('sep');
        helpers_fn_1.log(`willLint ${fileName}`, 'info');
        helpers_fn_1.log('sep');
        await lint_fn_1.lintFn(fileName, 'local', false, true);
    }
    onJestSuccess(input) {
        const { execResult, specFile, actualFileName, fileName, dir, extension, hasTypescript, } = input;
        const hasError = execResult.stderr.startsWith('FAIL') ||
            execResult.stderr.includes('ERROR:');
        const { pass, message, uncovered } = this.parseCoverage({
            execResult,
            hasError,
            actualFileName,
            fileName,
            extension,
        });
        const { newDecorations, hasDecorations } = get_new_decorations_1.getNewDecorations({
            execResult,
            fileName,
            hasTypescript,
        });
        const firstBarMessage = pass ? message : common_1.ERROR_ICON;
        const secondBarMessage = get_uncovered_message_1.getUncoveredMessage(uncovered);
        debugLog({
            pass,
            hasTypescript,
            message,
            hasError,
            newDecorations,
            hasDecorations,
        }, 'vscode.message');
        const shorterSpecFile = rambdax_1.remove(dir, specFile);
        this.emit({
            firstBarMessage,
            secondBarMessage,
            thirdBarMessage: shorterSpecFile,
            hasDecorations,
            newDecorations,
        });
    }
    async execJest(input) {
        const { fileName, dir, specFileName } = input;
        const [coveragePath, actualFileName, extension] = get_coverage_path_1.getCoveragePath(dir, fileName);
        try {
            const testPattern = `-- ${specFileName}`;
            const command = [
                common_1.JEST_BIN,
                '-u',
                '--detectOpenHandles',
                '--maxWorkers=1',
                '--env=node',
                '--collectCoverage=true',
                coveragePath,
                testPattern,
            ].join(' ');
            this.jestChild = execa_1.default.command(command, { cwd: dir });
            helpers_fn_1.log('sepx');
            helpers_fn_1.log('Jest start', 'info');
            const result = await this.jestChild;
            helpers_fn_1.log('Jest end', 'info');
            this.jestChild = undefined;
            return [false, result, actualFileName, extension];
        }
        catch (e) {
            if (e.isCanceled)
                return [true];
            if (!e.stdout)
                return [true];
            if (!e.stderr)
                return [true];
            return [
                false,
                {
                    stdout: e.stdout,
                    stderr: e.stderr,
                },
                actualFileName,
                extension,
            ];
        }
    }
    parseCoverage(parseCoverageInput) {
        const { execResult, actualFileName, fileName, extension, hasError, } = parseCoverageInput;
        if (hasError)
            return {};
        const input = common_1.cleanAngularLog(execResult);
        const pass = input.stderr.includes('PASS');
        const jestOutputLines = input.stdout.split('\n');
        let foundCoverage = false;
        const [lineWithCoverage] = jestOutputLines.filter(line => {
            if (line.includes('% Stmts'))
                foundCoverage = true;
            return foundCoverage && line.includes(`${actualFileName}${extension}`);
        });
        if (lineWithCoverage === undefined) {
            return {
                pass,
                message: common_1.SUCCESS_ICON,
            };
        }
        const [, statements, branch, func, lines, uncovered,] = lineWithCoverage.split('|').map(common_1.extractNumber);
        const message = this.getCoverageDiff([statements, branch, func, lines], fileName);
        return {
            pass,
            message: message === undefined ? 'MESSAGE === undefined' : message,
            uncovered,
        };
    }
    getCoverageDiff(inputs, filePath) {
        const fileKey = create_file_key_1.createFileKey(filePath);
        const firstTime = this.coverageHolder[fileKey] === undefined;
        const [statements, branch, func, lines] = inputs.map(common_1.toNumber);
        const hash = {
            branch,
            func,
            lines,
            statements,
        };
        const coverageInfo = rambdax_1.glue(`
    statements ${STATEMENTS}:
    ${statements} % ${SEPARATOR}
    branches ${BRANCHES}:
    ${branch} % ${SEPARATOR}
    functions ${FUNCTIONS}:
    ${func} % ${SEPARATOR}
    lines ${LINES}:
    ${lines} %
  `);
        if (firstTime) {
            this.coverageHolder[fileKey] = hash;
            return coverageInfo;
        }
        const statementsDiff = common_1.parse(statements - this.coverageHolder[fileKey].statements);
        const branchDiff = common_1.parse(branch - this.coverageHolder[fileKey].branch);
        const funcDiff = common_1.parse(func - this.coverageHolder[fileKey].func);
        const linesDiff = common_1.parse(lines - this.coverageHolder[fileKey].lines);
        this.coverageHolder[fileKey] = hash;
        const message = rambdax_1.glue(`
      ${statementsDiff === 0 ? '' : `✍:${common_1.maybeWarn(statementsDiff)}`}
      ${branchDiff === 0 ? '' : `🎋:${common_1.maybeWarn(branchDiff)}`}
      ${funcDiff === 0 ? '' : `☈:${common_1.maybeWarn(funcDiff)}`}
      ${linesDiff === 0 ? '' : `📜:${common_1.maybeWarn(linesDiff)}`}
    `);
        const staleCoverageInfo = rambdax_1.glue(`
    NO CHANGE ${STALE_SEPARATOR}
    ${STATEMENTS}:
    ${statements} % ${STALE_SEPARATOR}
    ${BRANCHES}:
    ${branch} % ${STALE_SEPARATOR}
    ${FUNCTIONS}:
    ${func} % ${STALE_SEPARATOR}
    ${LINES}:
    ${lines} %
  `);
        return message.trim() === '' ? staleCoverageInfo : `change: ${message}`;
    }
    onCancelMessage() {
        if (!this.jestChild)
            return helpers_fn_1.log('', 'error');
        if (!this.jestChild.cancel)
            return;
        this.jestChild.cancel();
        this.jestChild = undefined;
    }
    async onSocketData(messageFromVSCode) {
        let parsedMessage;
        try {
            parsedMessage = JSON.parse(messageFromVSCode.toString());
        }
        catch (_) {
            parsedMessage = false;
        }
        if (parsedMessage === false) {
            return onWrongIncomingMessage(messageFromVSCode.toString());
        }
        debugLog(parsedMessage, 'onSocketData');
        if (parsedMessage.requestCancelation) {
            return this.onCancelMessage();
        }
        await this.onJestMessage(parsedMessage);
    }
    start() {
        if (this.initialized)
            helpers_fn_1.log('Already initialized', 'box');
        this.server = net_1.createServer(socket => {
            this.initialized = true;
            socket.on('data', data => this.onSocketData(data.toString()));
            socket.on('error', err => {
                this.initialized = false;
                console.log(err, 'socket.error.niketa.client');
                this.server.close(() => {
                    rambdax_1.delay(2000).then(() => {
                        this.start();
                    });
                });
            });
            this.emit = message => {
                socket.write(JSON.stringify(message));
                socket.pipe(socket);
            };
        });
        this.server.listen(this.port, '127.0.0.1');
    }
}
exports.NiketaClient = NiketaClient;
